package jsystem;

import java.util.*;
import java.io.*;
import kjava.io.*;

import javax.microedition.io.*;
import javax.microedition.io.file.*;
import javax.microedition.lcdui.*;



public class FileBrowser extends List implements CommandListener{
	Display 			display;
	Displayable 		back, fb;		// Show when back button is pressed
	FileHandler[] 		filehandler;
	
  	private final static String[] typeList = { "Regular File", "Directory" };
    private Image 		dirIcon, fileIcon;
    private Image[] 	iconList;

  	private Form 		creator 	= new Form("New File");
	private TextField 	nameInput 	= new TextField("Enter Name", null, 256, TextField.ANY);
	private ChoiceGroup typeInput 	= new ChoiceGroup("Enter File Type", Choice.EXCLUSIVE,typeList, iconList);
  	private Alert 		accept		= new Alert(" "," ",null,null);
  	private TextBox 	filename	= new TextBox("","",256,TextField.ANY);

  	private Command 	okCmd      	= new Command("OK", Command.OK, 1);
  	private Command 	backCmd 	= new Command("Back", Command.BACK, 1);
  	private Command 	newCmd     	= new Command("New", Command.ITEM, 2);
  	private Command 	copyCmd     = new Command("Copy", Command.ITEM, 2);
  	private Command 	rmCmd      	= new Command("Delete", Command.ITEM, 2);
  	private Command 	openCmd    	= new Command("Open", Command.ITEM, 2);
  	private Command 	cancelCmd  	= new Command("Cancel", Command.CANCEL, 1);
  	private Command 	saveCmd    	= new Command("Save", Command.OK, 1);
  	private Command 	uploadCmd  	= new Command("Upload", Command.ITEM, 2);
  	private Command 	downloadCmd	= new Command("Download", Command.ITEM, 2);
  	private Command 	editnCmd   	= new Command("Rename", Command.ITEM, 2);


  	byte[] 				loadedFile 	= null;

    public String 		currDirName;
	String 				currFile 	= null, 
						delFile		= null;


    /* special string denotes upper directory */
    private final static String UP_DIRECTORY = "..";
    
    /* special string that denotes apper directory accessible by this browser.
     * this virtual directory contains all roots.
     */
    private final static String MEGA_ROOT = "/";
    
    /* separator string as defined by FC specification */
    private final static String SEP_STR = "/";
    
    /* separator character as defined by FC specification */
    private final static char   SEP = '/';
	


	public FileBrowser(Display display, Displayable back, FileHandler[] filehandler){
		super("File Browser", List.IMPLICIT);
		this.display = display;
		this.back = back;
		this.filehandler=filehandler;
		fb = this;

		setSelectCommand(openCmd);
		addCommand(newCmd);
		addCommand(rmCmd);
		addCommand(copyCmd);
		addCommand(backCmd);
		addCommand(uploadCmd);
		addCommand(downloadCmd);
		addCommand(editnCmd);
		setCommandListener(this);

        currDirName = MEGA_ROOT;
        try {
            dirIcon = Image.createImage("/icons/dir.png");
        } catch (IOException e) {
            dirIcon = null;
        }
        try {
            fileIcon = Image.createImage("/icons/file.png");
        } catch (IOException e) {
            fileIcon = null;
        }
        iconList = new Image[] { fileIcon, dirIcon };

     	accept.addCommand(okCmd);
     	accept.addCommand(cancelCmd);
     	accept.setCommandListener(this);

     	filename.addCommand(okCmd);
     	filename.addCommand(cancelCmd);
     	filename.setCommandListener(this);
		
		creator.append(nameInput);
		creator.append(typeInput);
		creator.addCommand(okCmd);
		creator.addCommand(backCmd);
		creator.setCommandListener(this);		
	}
	
  	private String tname="";	
	
	public void commandAction(Command c,
                          Displayable d){
		if(d == this){
            currFile = getString(getSelectedIndex());
      		if( c == backCmd) {
          		display.setCurrent(back);
         	}else if( c == newCmd) {
		 		display.setCurrent(creator);
         	}else if( c == openCmd) {
				if(currFile==null)
             		JSystem.err("Nothing selected for open");
				else
					new Thread(new Runnable() {
                		public void run() {
                    		if (currFile.endsWith(SEP_STR) || currFile.equals(UP_DIRECTORY)) {
                        		traverseDirectory(currFile);
                    		} else 
								openFile(currDirName + currFile);
                		}
            		}).start();
			}else if( c == copyCmd) {
				if(currFile==null){
                      JSystem.err("Nothing selected for Copying");
				}else if (currFile.endsWith(SEP_STR) || currFile.equals(UP_DIRECTORY)) {
                      JSystem.err("Can not copy Directory"); 
				}else{
            		new Thread(new Runnable() {
                		public void run() {
							byte[] b = readFile(currDirName + currFile);
							if(b!=null){
								loadedFile = b;
								nameInput.setString( currFile );
							}
                		}
            		}).start();
				}
          	}else if( c == rmCmd) {
            	if(currFile==null){ // Nothing 
             		JSystem.err("Nothing selected for delete");
             	}else{
					delFile = currDirName + currFile;
            		accept.setString("Really delete file "+delFile+"?");
             		display.setCurrent(accept);
          		}
          	}else if( c == editnCmd) {
          		if(currFile==null){ // Nothing 
             		JSystem.err("Nothing selected");
     			}else if(currFile.equals(UP_DIRECTORY) || currFile.equals(MEGA_ROOT)){ 
				 	JSystem.err("Can't change names for this directory");
     			}else if(currDirName.startsWith("res/")){ 
				 	JSystem.err("Read-only file system.");
     			}else{
               		filename.setTitle("Edit Filename");
            		filename.setString(currFile);
             		tname=currFile;
       				display.setCurrent(filename);
     			}
          	}
		}else if(d==filename){
     		if( c == okCmd ) {
     			if(filename.getString()!=null &&filename.getString().length()!=0
                     && !tname.equals(filename.getString())){
					// Rename file/directory
					new Thread(new Runnable() {
            			public void run() {
							if("vfs/".equals(currDirName)){
								try{
									JSystem.vfs.rename(tname, filename.getString());
								} catch (Exception e) {
									JSystem.err("Error renaming file on vfs.");
								}
							}else{
       							try {
            						FileConnection fc = (FileConnection) Connector.open(
										"file://localhost/" + currDirName + tname);
									if(fc.exists())
										fc.rename(filename.getString());
									fc.close();
        						} catch (Exception e) {
									JSystem.err("Error renaming file.");
								}
							}
							filename.setString("");   
							updateDirlist();
               				display.setCurrent(fb);      		
               			}
					}).start();
				}            
     		} if( c==cancelCmd ){
     			display.setCurrent(this);
     			filename.setString("");
  			}
		}else if(d==accept){
         	if( c == okCmd) {
             	if(accept.getString().startsWith("Really delete file")){
            		new Thread(new Runnable() {
                		public void run() {
       						if (delFile.equals(UP_DIRECTORY)) {
                        		;
                    		} else if(delFile.startsWith("res/") ){
								JSystem.err("Read-only file system.");
							}else {
                        		// Delete file/directory
								if(delFile.startsWith("vfs/")){
									if(delFile.equals("vfs/")){
										JSystem.err("vfs can not be removed");
									}else{
										try{
											JSystem.vfs.delete(delFile.substring("vfs/".length()));
										} catch (Exception e) {
											JSystem.err("Error deleting file on VFS.");
										}
									}
								}else{
									try {
            							FileConnection fc = (FileConnection) Connector.open(
											"file://localhost/" + delFile);
										if(fc.exists())
											fc.delete();
										fc.close();
        							} catch (Exception e) {
										JSystem.err("Error deleting file.");
									}
								}
                			}
							updateDirlist();
                  			display.setCurrent(fb);
            			}
					}).start();
				}
        	}else if(c==cancelCmd){
				if(accept.getString().startsWith("Really delete file"))
             		display.setCurrent(this);
        	}
     	}else if (d == creator){
			if (c == backCmd) {
				new Thread(new Runnable() {
          			public void run() {
         				updateDirlist();
            			display.setCurrent(fb); 
                	}
            	}).start();
			}else if (c == okCmd) {
            	final String newName = nameInput.getString();
            	if (newName == null || newName.equals("")) {
            		JSystem.err("File Name is empty. Please provide file name");
            	} else {
					new Thread(new Runnable(){
            			public void run(){
                			createFile(newName, typeInput.getSelectedIndex() != 0);
            			}
        			}).start();
         		}
			}
		}

	}

    void createFile(String newName, boolean isDirectory) {  
		if(currDirName.startsWith("res/")){
			JSystem.err("Resourses are read only");
		}else if("vfs/".equals(currDirName)){
			if(isDirectory)
				JSystem.err("No directories on virtual file system.");
			else{
				try{ 
					if( loadedFile != null){
						Vfs.write(loadedFile, loadedFile.length, newName);
						loadedFile = null;
					}else
						Vfs.write(new byte[0], 0, newName);
				}catch(Exception e){
					JSystem.err("Error creating file on VFS");
				}
			}
		}else{
        	try {
            	FileConnection fc = (FileConnection) Connector.open("file:///" +
            			currDirName + newName);
            	if (isDirectory) {
                	fc.mkdir();
            	} else {
                	fc.create();
					if( loadedFile != null){
						OutputStream os = fc.openOutputStream();
						os.write(loadedFile);
						loadedFile = null;
						os.close();
					}
				}
				fc.close();
            }catch (Exception e) {
					JSystem.err("Error creating file/directory.");
         	}
		}
		updateDirlist();
        display.setCurrent(this); 
    }




	void traverseDirectory(String fileName) {
    /* In case of directory just change the current directory
     * and show it
     */
		if (currDirName.equals(MEGA_ROOT)) {
            if (fileName.equals(UP_DIRECTORY)) {
                // can not go up from MEGA_ROOT
                return;
            }
            currDirName = fileName;
        } else if (fileName.equals(UP_DIRECTORY)) {
            // Go up one directory
            // TODO use setFileConnection when implemented
            int i = currDirName.lastIndexOf(SEP, currDirName.length()-2);
            if (i != -1) {
                currDirName = currDirName.substring(0, i+1);
            } else {
                currDirName = MEGA_ROOT;
            }
        } else {
            currDirName = currDirName + fileName;
        }
		updateDirlist();
    }




	
	public void updateDirlist(){
    	deleteAll();
 		Enumeration e=null;
    	try {
   			if (MEGA_ROOT.equals(currDirName)) {
    			e = FileSystemRegistry.listRoots();
			}else if("vfs/".equals(currDirName)){
				e = JSystem.vfs.getFileList();
        		append(UP_DIRECTORY, dirIcon);
			}else if(currDirName.startsWith("res/") ){
				// Read directory listing from File dir
				Vector flist = new Vector();
				try{				
					BufferedReader br = new BufferedReader(new FileReader(flist.getClass().getResourceAsStream(
								 		currDirName.substring("res".length())+".dir" )));
					String p;
					while( (p=br.readLine()) != null){
						flist.addElement(p);
					}
					br.close();
				}catch(Exception e2){
				}
				e = flist.elements();
        		append(UP_DIRECTORY, dirIcon);
   			} else {
  				FileConnection currDir =  (FileConnection)Connector.open("file://localhost/" + 
                            currDirName);
    			e = currDir.list();
				currDir.close();
        		append(UP_DIRECTORY, dirIcon);
    		}
			setTitle(currDirName);
		
  			while (e!=null && e.hasMoreElements()) {
        		String fileName = (String)e.nextElement();
       			if (fileName.charAt(fileName.length()-1) == SEP) {
                    // This is directory
          			append(fileName, dirIcon);
  				} else {
                    // this is regular file
           			append(fileName, fileIcon);
        		}
    		}

			if (MEGA_ROOT.equals(currDirName)){
				append("vfs/", dirIcon);
				append("res/", dirIcon);
			}
			
        } catch (IOException ioe) {
            ioe.printStackTrace();
        }
	}
  	
	void openFile(String fname){
		if(filehandler != null){
			for(int i=0; i<filehandler.length; i++){
				FileHandler f=filehandler[i];
				if(f.canHandle(fname)){
					f.start(display, back, fname);
					return;
				}
			}
		}
		JSystem.err("No registered Filehandler for "+fname);
	}	
	
	public static byte[] readFile(String fname){
		try{
			int len = 0;
			if(fname.startsWith("res/")){
				FileInputStream f = new FileInputStream(fname);
				// Count length
				while( f.read() != -1 ){
					len++;
				}
				f.close();
			}else{
				len = (int)(new File(fname)).length();
			}
			if(len>0){
				byte b[] = new byte[len];
				if(b==null){
					return null;
				}
				FileInputStream	f = new FileInputStream(fname);
				int pos = 0;
				while(pos<len){
					int r = f.read(b,pos,len-pos);
					if(r<0) return null;
					pos += r;
				}
            	f.close();
				return b;
			}
			return null;
		}catch(Exception e){
			return null;
		}
	}
}
